#include <cctype>
#include <cstdio>
#include <cstdarg>
#include <errno.h>
#include <fcntl.h>
#include <iostream>
#include <unistd.h>
#include <sys/stat.h>

#define TELOPTS
#define TELCMDS
#include <arpa/telnet.h>

#include "playertask.hh"
#include "taskmanager.hh"
#include "stringfun.hh"
#include "events.hh"

////////////

static const bool BROKEN_TERM = false;
static const bool SUPPORT_2x  = true;
static const bool COLOURS     = true;

class playerdata
{
	playertask &owner;
	unsigned char attr;	
    void colour(unsigned char newattr)
    {
    	if(newattr == attr || !COLOURS)return;
    	
    	string Buf = "\33[";
    	
    	if(BROKEN_TERM || newattr != 7)
    	{
    		static const char Swap[] = "04261537";
    		bool pp;
    		#define P(x) do{if(pp)Buf+=';';Buf += x;pp=true;}while(0)
    		#define P2(a,b) do{if(pp)Buf+=';';Buf+=a;Buf+=b;pp=true;}while(0)
    		if(SUPPORT_2x)
    		{
    			if((attr & 0x88) && !(newattr & 0x88)) P('0');
    			else if((attr & 0x80) && !(newattr & 0x80)) P("25");
    			else if((attr & 0x08) && !(newattr & 0x08)) P("21");
    			else if((newattr & 0x80) && !(attr & 0x80)) P('5');
    			else if((newattr & 0x08) && !(attr & 0x08)) P('1');
    		}
    		else
    		{
    			if(((attr & 0x80) && !(newattr & 0x80))
    			|| ((attr & 0x08) && !(newattr & 0x08)))
    			{
    				P("0");
    				newattr = BROKEN_TERM ? (attr&0x77) ^ 0x77 : 7;
    			}
    			if((newattr & 0x80) && !(attr & 0x80)) P('5');
    			if((newattr & 0x08) && !(attr & 0x08)) P('1');
    		}
    		if((newattr & 0x70) != (attr & 0x70)) P2('4', Swap[(newattr>>4)&7]);
    		if((newattr & 0x07) != (attr & 0x07)) P2('3', Swap[newattr&7]);
    		#undef P
    	}
    	Buf += 'm';
    	owner.echo(Buf, false);
   		attr = newattr;
    }
    
    void displayprompt()
    {
    	litecho(prompt);
    }

public:
	enum
	{
		// Never add new situations in the
		// middle, always add them to the end!
		S_LOGIN = 0,
		S_GAME
	} situation;
	
	string name;
	string prompt;
	
public:
	playerdata(playertask &o) : owner(o), attr(0x07), situation(S_LOGIN)
	{
		prompt = "`D>";
	
    	litecho(
    	    "`F---------------------------------------------------------------------------\n"
    	    "`A*** `FDUD4 `Aruntime - copyright (C) `B1992`A,`B2000`A Bisqwit (`Bhttp://iki.fi/bisqwit/`A)\n"
    	    "`F---------------------------------------------------------------------------\n"
    	    "`3DUD login v"VERSION" - recommended to have square font.\n"
    	    "`7\n"
    	    "Enter your character name (one word, case insignificant): ");
	}
	
	~playerdata()
	{
		save();
	}
	
	void load()
	{
		FILE *fp = fopen(("players/" + name).c_str(), "rb");
		if(fp)
		{
			fclose(fp);
		}
		else
		{
			situation = S_GAME;
		}
	}
	
	void save()
	{
		
	}

    void telnetsend(int dummy, ...)
    {
    	va_list ap;
    	va_start(ap, dummy);
    	
    	int a;
    	string Buf;
    	
    	Buf += (char)dummy;
    	
    	while((a = va_arg(ap, int)) >= 0)
    		Buf += (char)a;
    		
    	owner.echo(Buf, false);
    }

    void noecho()
    {
    	telnetsend(IAC, WONT, TELOPT_ECHO, -1);
    }

    void doecho()
    {
    	telnetsend(IAC, WILL, TELOPT_ECHO, -1);
    }
    
    void caninput()
    {
		telnetsend(IAC, GA, -1);
		colour(7);
    }
    
    void litecho(const string &s)
    {
    	const char *q = s.c_str();
    	const char *p;
    	
    	for(;;)
    	{
    		p = strchr(q, '`');
    		if(!p)
    		{
    			owner.echo(q, false);
    			return;
    		}
    		owner.echo(string(q, 0, p-q), false);
    		q = p + 2;
    		if(!p[1])break;
    		if(p[1] == '`')owner.echo("`", false);
    		else if((p[1] >= '0' && p[1] <= '9'))
    			colour((attr&0xF0) | (p[1]-'0'));
    		else if((p[1] >= 'A' && p[1] <= 'F'))
    			colour((attr&0xF0) | (p[1]-'A'+10));
    	}
    }
    
    void handleinput()
	{
		vector<string> args;
		string input = owner.input;
		
		switch(situation)
		{
			case S_LOGIN:
				name = strtolower(argsplit(args, input));
				load();
				break;
			default:
				;
		}
		
		displayprompt();
	}
	
	void tick()
	{
	}
};


playertask::playertask(int s) : sock(s), lineterm(0), got_iac(0)
{
	data = new playerdata(*this);
	
	data->telnetsend(
		IAC, DO, TELOPT_BINARY,
		-1);
	
	cout << "[" << sock << "]\n";
	
	data->caninput();
}

playertask::~playertask()
{
	cout << "[~" << sock << "]\n";
	
	delete data;
	close(sock);
}

void playertask::echo(const string &s, bool endcr)
{
	fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) &~ O_NONBLOCK);
	
	const char *q = s.c_str();
	unsigned c = s.size();
	
	while(c)
	{
		int n = write(sock, q, c);
		if(n < 0)break;
		q += n;
		c -= n;			
	}
	
	if(endcr)
	    write(sock, "\r\n", 2);
	
	fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
}

bool playertask::tick()
{
	data->tick();
	
	unsigned char ch;
	int n;
	while((n = read(sock, &ch, 1)) == 1)
	{
		switch(got_iac)
		{
			case 0:
			{
				if(ch == IAC)
					++got_iac;
				else
				{
            		switch(ch)
            		{
            			case '':
            			case '':
            				if(input.size() > 0)input = input.substr(0, input.size()-1);
            				break;
            			case '\r':
            			case '\n':
            				if(lineterm && lineterm!=ch)break;
            				lineterm = ch;
            				
							cout << "[" << sock << "]: " << input << endl;
							
            				data->handleinput();
            				input = "";
            				data->caninput();
            				break;
            			default:
            				if(ch >= ' ')input += ch;
            		}
            	}
				break;
			}
			case 1: // got_iac == 1
				switch(ch)
				{
					case DO:
					case DONT:
					case WILL:
					case WONT:
						got_iac = ch;
						break;
					case IP:
						input = "";
						// fall-through
						echo("^C");
					case AYT:
						// fall-through
					default:
						cout << "Got: IAC " << telcmds[ch-xEOF] << endl;
						got_iac = 0;
						break;
				}
				break;
			default:
				switch(ch)
				{
					default:
						if(got_iac == DO)  data->telnetsend(IAC, WILL, ch, -1);
						if(got_iac == DONT)data->telnetsend(IAC, WONT, ch, -1);
						cout << "Got: IAC " << telcmds[got_iac-xEOF] << " " << telopts[ch] << endl;
				}
				got_iac = 0;
		}
	}
	if(!n)return false; /* Socket died */
	if(n < 0 && errno != EAGAIN && errno != EINTR)
	{
	//	extern const char *sys_errlist[];
		cerr << "[!" << sock << "]: " << sys_errlist[errno] << endl;
		return false;
	}
	return true;
}
